/*
	rewings (copyleft) 2004 by Moritz Voss

	 -  repacks WINGSPAK files for Sldner: Secret Wars
	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	
*/


/* 
  This program requires zlib version 1.2.1 or highter
  zlib is Copyright (C) 1995-2003 Jean-loup Gailly and Mark Adler
*/

#define  ZLIB_WINAPI 

#include <queue>

#include <fstream>
#include <iostream>

#include <zlib.h>   //good old zlib
#include <direct.h> //for directory manipulation

using namespace std;

#define WINGSPAK_HEADER_LENGTH 12

#define OPTION_OUTPUT_SILENT		0
#define OPTION_OUTPUT_NORMAL		1
#define OPTION_OUTPUT_VERBOSE		2

#define OPTION_RECURSE				1


#define OPTION_MODE_LIST	0
#define OPTION_MODE_PACK	1

#define OPTION_WALK_RECURSE 1


struct tOptions
{ 
	unsigned int output;	//0 silent, 1 normal, 2 verbose
	unsigned int mode;		//0 LIST, 1 pack
	
	unsigned int walk;
	int compression; // 0, 1..9


	bool force;	
};



std::ofstream pak; // the wingspak file

tOptions options;


void writeBigEndian(unsigned long dword) //write unsigned long to file in BIG ENDIAN (arrrrrgh)
{
	pak.put((char) (dword >> 24 & 255));	
	pak.put((char) (dword >> 16 & 255));	
	pak.put((char) (dword >> 8 & 255));	
	pak.put((char) (dword & 255));			
}



class cWPItem
{
public:
	unsigned long run_length;
	unsigned long file_offset;
	unsigned long file_size;
	unsigned char is_directory;
	unsigned char string_length;
	char* name_string; 

	cWPItem(const char* name, bool directory, unsigned long size) //instantiate new file on disk
	{
		name_string = new char[strlen(name)+1];
		memcpy(name_string, name, strlen(name)+1);
		is_directory = directory;
		file_size = size;
	}

	~cWPItem(void)
	{
		delete name_string;
	}

	void store() //compress a file using zlib and store to disk, reading from global stream
	{
		std::ifstream infile;
		void* data;
		void* dest;
		unsigned long dest_length;

		infile.open(name_string, ios_base::binary);
		data = new char[file_size];
		dest = new char[dest_length = file_size+12+file_size/900]; //900 instead of 1000 so we dont get a silly +1 error 2 years after release or so

		infile.read((char*) data, file_size);
		infile.close();

		if (int zerror = compress2((Bytef *)dest, &dest_length, (Bytef *)data, file_size, options.compression))
			if (options.output) cout << "Error: zlib returned code " << zerror << ".\n";
		
		file_offset = pak.tellp(); //note the offset

		pak.write((char*)&file_size, 4); //store uncompressed size of file as little endian unsigned long

		file_size = dest_length + 4;


		pak.write((char*)dest, dest_length); //finally, write the uncompressed stream

		delete data; //cleanup
		delete dest;
	}

};



std::queue<cWPItem*> table;  //actual file table in memory before writing



unsigned long rePakChunk(const char* filename, bool root)
{
	cWPItem* item;
	unsigned long current_sub_chunks = 0;


	HANDLE handle;
	WIN32_FIND_DATA data; //don't really like it, but for now it'll work

	if ((handle = FindFirstFile("*.*",&data)) == INVALID_HANDLE_VALUE) 
	{
		return 0;
	}

	// loop finding all files that matter
	while (FindNextFile(handle,&data))
	if (strcmp(data.cFileName, "..") && !(root && !strcmp(data.cFileName, filename))
		&& (options.walk || data.dwFileAttributes != FILE_ATTRIBUTE_DIRECTORY)) //don't ascend here in given cases
		{
			current_sub_chunks++;
			item = new cWPItem(data.cFileName, data.dwFileAttributes == FILE_ATTRIBUTE_DIRECTORY, data.nFileSizeLow); //4 gb max :-/

			table.push(item); //store in temporary table

			if (item->is_directory)
			{
				switch (options.output)
				{
					case OPTION_OUTPUT_SILENT: break;
					case OPTION_OUTPUT_VERBOSE:
						cout << "DIRECTORY : " << data.cFileName << "\n";
						break;
					default:
						cout << "D " << data.cFileName << "\n";
				}

				chdir(item->name_string);
				item->run_length = rePakChunk(filename, false); //store number of subchunks
				chdir(".."); //leave the dir again
			} 
			else 
			{	
				switch (options.output)
				{
					case OPTION_OUTPUT_SILENT: break;
					case OPTION_OUTPUT_VERBOSE:
						cout << "FILE : " << data.cFileName << " : (" << data.nFileSizeLow << " bytes)\n";
						break;
					default:
						cout << "F " << data.cFileName << "\n";
				}
				item->run_length = 1; //normal files have weight 1
				item->store();
			}
		}

	if (GetLastError() != ERROR_NO_MORE_FILES) 
	{
		cout << "Error: File system returned code " << GetLastError() << ".\n";
	}
	FindClose(handle);
	
	return current_sub_chunks; // for recursion and all
}

void rePak(const char*filename, const char* directory)
{
					pak.open(filename, ios_base::binary);

					pak.write("WINGSPAK", strlen("WINGSPAK"));

					writeBigEndian(0);
					writeBigEndian(0); //for the address later
					writeBigEndian(0); //for the chunks later


					unsigned long chunks = rePakChunk(filename, true);

					unsigned long table_offset = pak.tellp(); //determine offset of file allocation table before writing it

					writeBigEndian(chunks); //yep dont forget da chunxxxx

					while (!table.empty()) //purge table and write its contents to disk
					{
						pak.put((char) strlen((table.front())->name_string));//write pascal style strlen
						pak.write((table.front())->name_string, strlen((table.front())->name_string));

						pak.put((table.front())->is_directory); //directory flag

						if (!((table.front())->is_directory))
						{
							writeBigEndian(0); //too lazy for 64 bit... this is a computer game, after all
							writeBigEndian((table.front())->file_offset);

							writeBigEndian(0);
							writeBigEndian((table.front())->file_size);
						}
						writeBigEndian((table.front())->run_length);

						delete table.front(); //get rid of the sucker
						table.pop(); //next!
					}

					pak.seekp(WINGSPAK_HEADER_LENGTH);
					writeBigEndian(table_offset); //finalize. yessss!
					pak.close();
}

void beginPak(const char* filename, const char* directory)
{
	if (chdir(directory))
	{
		if (options.output) cout << "Error: Invalid directory.\n";
	}
	else //do the walking and writing
	{
		if (!options.force)
		{
			std::ifstream testfile;
			testfile.open(filename); //check with a tentative input open

			if (!testfile) //doesn't exist
			{
				rePak(filename, directory);
			} else
			{
				testfile.close();
				char input;
				cout << filename << " already exists. Overwrite? (y/n/a):";
				cin >> input;
				if (input == 'y' || input == 'a')
				{
					if (input == 'a') options.force = true;
					rePak(filename, directory);
				}
			}
		}
	}
}


//display a little help screen
void doHelp(void)
{
	cout << "rewings 0.0.1.0 \n"
			 << "  A tool that creates compressed WINGSPAK files using zlib.\n"
			 << "  (c)2004 by Moritz Voss (public@thygrrr.de)\n"
			 << "\n"
			 << "  This program is free software; you can redistribute it and/or\n"
			 << "  modify it under the terms of the GNU General Public License.\n"
			 << "\n"
			 << "Usage:\n"
			 << "  rewings -<options> <pak-file> [target-dir]\n"
			 << "\n"
			 << "Options:\n"
			 << "  c : compress files into <pak-file> (default: current directory)\n"
			 << "  l : only list files, don't actually write to disk\n"
			 << "  r : recurse into subdirectories\n"
			 << "  f : assume 'yes' on all questions (e.g. overwrite file?)\n"
			 << "  s : suppress all output except interactive warnings\n"
			 << "  n : normal output (default)\n"
			 << "  v : verbose output\n"
			 << "0..9: compression - 0=none, 1=fast, 9=best (default: 6)\n"
			 << "\n"
			 << "Examples:\n"
			 << "  rewings -c mysoldner.pak modstuff \n"
			 << "   --> compresses every file in directory modstuff into mysoldner.pak\n"
			 << "  rewings -cr9 mysoldner.pak\n"
			 << "   --> compresses current directory, files, and all subdirectories with\n"
			 << "       best compression ratio into the file called mysoldner.pak\n"
			 << "  rewings -lv mysoldner.pak modstuff \n"
			 << "   --> verbosely lists the files and steps of compression, but writes\n"
			 << "       no data to disk\n";

	exit(0); //shutdown program
}




// main function, just parse the command line parameters, then call actual worker functions
int main( int argc, char *argv[ ] )
{
	switch (argc)
	{
		case 3: 
		case 4:
			{
				if (*argv[1] == '-') //it's a valid options string
				{
					options.output = OPTION_OUTPUT_NORMAL; //default setting here
					options.compression = Z_DEFAULT_COMPRESSION;


					//determine wanted compression / speed ratio
					if (strchr(argv[1], '0')) options.compression = 0;
					if (strchr(argv[1], '1')) options.compression = 1;
					if (strchr(argv[1], '2')) options.compression = 2;
					if (strchr(argv[1], '3')) options.compression = 3;
					if (strchr(argv[1], '4')) options.compression = 4;
					if (strchr(argv[1], '5')) options.compression = 5;
					if (strchr(argv[1], '6')) options.compression = 6;
					if (strchr(argv[1], '7')) options.compression = 7;
					if (strchr(argv[1], '8')) options.compression = 8;
					if (strchr(argv[1], '9')) options.compression = 9;
					
						
					strchr(argv[1], 'r') ? options.walk = OPTION_WALK_RECURSE : options.walk = 0;


					if (strstr(argv[1], "c")) //user wants to pack
					{
						if (strchr(argv[1], 'l')) doHelp(); //c, l and x are mutually exclusive
						options.mode = OPTION_MODE_PACK;
					}

					if (strstr(argv[1], "l")) //user wants to list
					{
						if (strchr(argv[1], 'c')) doHelp(); //c, l and x are mutually exclusive
						options.mode = OPTION_MODE_LIST;
					}

					if (strchr(argv[1], 's')) //user wants to list
					{
						if (strchr(argv[1], 'v')) doHelp(); //v and s are mutually exclusive
						options.output = OPTION_OUTPUT_SILENT;
					}
					
					if (strchr(argv[1], 'v')) //user wants to list
					{
						if (strchr(argv[1], 's')) doHelp(); //v and s are mutually exclusive
						options.output = OPTION_OUTPUT_VERBOSE;
					}

					options.force = (strchr(argv[1], 'f') != NULL); //force overwrite of existing files
					switch (options.mode)
					{
					default:
						if (argc == 4) beginPak(argv[2], argv[3]);
						else beginPak(argv[2], ".");
					}
				}
				else doHelp();
			}
			break;

		default: 
			doHelp();
	};

	return 0;
}
